"""
==============================================================
基于 Kernel LDA + KNN 的人脸识别
使用 Kernel Discriminant Analysis 做特征降维
使用 K-Nearest-Neighbor 做分类

数据:
    人脸图像来自于 Olivetti faces data-set from AT&T (classification)
    数据集包含 40 个人的人脸图像, 每个人都有 10 张图像
    我们只使用其中标签(label/target)为 0 和 1 的前 2 个人的图像

算法:
    需要自己实现基于 RBF Kernel 的 Kernel Discriminant Analysis 用于处理两个类别的数据的特征降维
    代码的框架已经给出, 需要学生自己补充 KernelDiscriminantAnalysis 的 fit() 和 transform() 函数的内容
==============================================================
"""

import numpy as np
import matplotlib.pyplot as plt
from sklearn import datasets
from sklearn.model_selection import train_test_split
from sklearn.datasets import fetch_olivetti_faces
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

print(__doc__)
################################################
"""
Scikit-learn-compatible Kernel Discriminant Analysis.
"""

import numpy as np
from scipy import linalg
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

class KernelDiscriminantAnalysis(BaseEstimator, ClassifierMixin,
                                 TransformerMixin):
    """Kernel Discriminant Analysis.

    Parameters
    ----------
    n_components: integer.
                  The dimension after transform.
    gamma: float.
           Parameter to RBF Kernel
    lmb: float (>= 0.0), default=0.001.
         Regularization parameter

    """

    def __init__(self, n_components, gamma, lmb=0.001):
        self.n_components = n_components
        self.gamma = gamma
        self.lmb = lmb
        self.X = None # 用于存放输入的训练数据的 X
        self.K = None # 用于存放训练数据 X 产生的 Kernel Matrix
        self.M = None # 用于存放 Kernel LDA 最优化公式中的 M
        self.N = None # 用于存放 Kernel LDA 最优化公式中的 N
        self.EigenVectors = None # 用于存放 Kernel LDA 最优化公式中的 M 对应的广义特征向量, 每一列为一个特征向量, 按照对应特征值大小排序

    def rbfkernel(self, gamma, distance):
        return np.exp(gamma * distance)

    def kernel(self, X, X_c, gamma):
        return np.exp(-gamma * np.sum((X - X_c) ** 2))

    def fit(self, X, y):
        """Fit KDA model.
        Parameters
        ----------
        X: numpy array of shape [n_samples, n_features]
           Training set.
        y: numpy array of shape [n_samples]
           Target values. Only works for 2 classes with label/target 0 and 1
        Returns
        -------
        self
        """
        self.X = X
        """
        distance_matrix = self.find_distance_matrix(X)
        K = self.rbfkernel(self.gamma, distance_matrix)
        self.K = K

        index1 = []
        index2 = []
        for i in range(len(y)):
            if y[i] == 0:
                index1.append(i)
            else:
                index2.append(i)

        K1 = []
        K2 = []
        # 遍历K矩阵的每一行
        for i in K:
            temp1 = []
            temp2 = []
            # 分别生成两类k_one的每一行
            for j in index1:
                temp1.append(i[j])
            for j in index2:
                temp2.append(i[j])
            # 把每一行都append进来
            K1.append(np.array(temp1))
            K2.append(np.array(temp2))
        K1 = np.array(K1)
        K2 = np.array(K2)

        l1 = len(index1)
        l2 = len(index2)
        A1 = np.identity(l1) - ((1 / float(l1)) * np.ones((l1, l1)))
        A2 = np.identity(l2) - ((1 / float(l2)) * np.ones((l2, l2)))

        N1 = np.dot(A1, K1.T)
        N1 = np.dot(K1, N1)

        N2 = np.dot(A2, K2.T)
        N2 = np.dot(K2, N2)

        self.N = N1 + N2 + self.lmb * np.identity(len(N1))
        M1 = []
        M2 = []
        for i in range(len(K1)):
            M1.append(np.sum(K1[i]) / float(l1))
        for i in range(len(K2)):
            M2.append(np.sum(K2[i]) / float(l2))
        M1 = np.array(M1)
        M2 = np.array(M2)

        M_diff = M1 - M2
        self.M = np.outer(M_diff.transpose(), M_diff)

        _, vecs = linalg.eig(self.M, self.N)
        #__, vecs = linalg.eig(np.linalg.inv(self.N).dot(self.M))
        self.EigenVectors = vecs

    def find_distance_matrix(self, data):
        euclid_distance = []
        for i in data:
            distance = []
            for j in data:
                distance.append(np.linalg.norm(i - j) * np.linalg.norm(i - j))
            distance = np.array(distance)
            euclid_distance.append(distance)
        euclid_distance = np.array(euclid_distance)
        return euclid_distance
        """

        
        #计算M矩阵
        def kernel_M(X, c=0):
            K_m = []
            c_len = len([i for i in y if i == c])
            for row in X:
                K_one = 0.0
                for c_row in X[y == c]:
                    K_one += self.kernel(row, c_row, self.gamma)
                K_m.append(K_one / c_len)
            return np.array(K_m)

        K = [np.zeros((X.shape[0], 1))] * 2
        for i in np.unique(y):
            K[i] = kernel_M(X, i)
        K_m = (K[0] - K[1])[:, np.newaxis].dot((K[0] - K[1])[np.newaxis, :])
        self.M = K_m

        # 计算N矩阵
        def kernel_N(X, c):
            c_len = len([i for i in y if i == c])
            I = np.eye(c_len)
            I_n = np.eye(N)
            I_c = np.ones((c_len, c_len)) / c_len
            K_one = np.zeros((N, c_len))

            for i in range(N):
                K_one[i, :] = np.array([self.kernel(X[i], c_row, self.gamma) for c_row in X[y == c]])
            K_n = K_one.dot(I - I_c).dot(K_one.T) + I_n * self.lmb  ##+ I_n*0.001
            return K_n

        N = X.shape[0]
        K_n = np.zeros((N, N))
        for i in np.unique(y):
            K_n += kernel_N(X, i)
        self.N = K_n
        # 计算特征值特征向量

        eigvals_, eigvecs_ = linalg.eig(self.M, self.N)
        eigvals, eigvecs = np.linalg.eig(np.linalg.inv(K_n).dot(K_m))
        eigen_pairs = [(np.abs(eigvals[i]), eigvecs[:, i]) for i in range(len(eigvals))]
        eigen_pairs = sorted(eigen_pairs, key=lambda k: k[0], reverse=True)
        alphas1 = eigen_pairs[0][1][:, np.newaxis]
        alphas2 = eigen_pairs[1][1][:, np.newaxis]
        EigenVectors = np.column_stack((alphas1, alphas2))
        self.EigenVectors = EigenVectors

    def transform(self, X_test):
        """Transform data with the trained KernelLDA model.

        Parameters
        ----------
        X_test: numpy array of shape [n_samples, n_features]
           The input data.

        Returns
        -------
        y_pred: array-like, shape (n_samples, n_components)
                Transformations for X.

        """
        """
        test_kernel_matrix = []
        for test_item in X_test:
            dist = np.array([np.sum((test_item - row) ** 2) for row in self.X])
            k = np.exp(-self.gamma * dist)
            test_kernel_matrix.append(k)

        y_pred = np.zeros((len(X_test), self.n_components))
        for i in range(0, self.n_components):
            alpha_i = self.EigenVectors[:, i]
            for j in range(len(test_kernel_matrix)):
                test_kernel = test_kernel_matrix[j]
                temp = 0
                for k in range(len(test_kernel)):
                    temp += alpha_i[k] * test_kernel[k]
                y_pred[j, i] = temp
        return y_pred

        """

        def project_x(X_new, X, ii):
            N = X_new.shape[0]
            X_proj = np.zeros((N, 1))
            for i in range(len(X_new)):
                k = np.exp(-self.gamma * np.array([np.sum((X_new[i] - row) ** 2) for row in X]))
                X_proj[i, 0] = np.real(k[np.newaxis, :].dot(self.EigenVectors[:, ii]))  ##不能带虚部
            return X_proj

        # 计算投影之后的点
        X_new = np.zeros((X_test.shape[0], 2))
        for i in range(self.n_components):
            X_new[:, i][:, np.newaxis] = project_x(X_test, self.X, i)  # alphas_one,最佳参数gamma=14.52

        return X_new

################################################

# 指定 KNN 中最近邻的个数 (k 的值)
n_neighbors = 3

# 设置随机数种子让实验可以复现
random_state = 0

# 现在人脸数据集
faces = fetch_olivetti_faces()
targets = faces.target

# show sample images
images = faces.images[targets < 2] # save images

features = faces.data  # features
targets = faces.target # targets

fig = plt.figure() # create a new figure window
for i in range(20): # display 20 images
    # subplot : 4 rows and 5 columns
    img_grid = fig.add_subplot(4, 5, i+1)
    # plot features as image
    img_grid.imshow(images[i], cmap='gray')

plt.show()

# Prepare data, 只限于处理类别 0 和 1 的人脸
X, y = faces.data[targets < 2], faces.target[targets < 2]

# Split into train/test
X_train, X_test, y_train, y_test = \
    train_test_split(X, y, test_size=0.5, stratify=y,
                     random_state=random_state)


# Reduce dimension to 2 with KernelDiscriminantAnalysis
# can adjust the value of 'gamma' as needed.
# why do we standarScaler data?
kda = make_pipeline(StandardScaler(),
                    KernelDiscriminantAnalysis(n_components=2, gamma=0.000005))

# Use a nearest neighbor classifier to evaluate the methods
knn = KNeighborsClassifier(n_neighbors=n_neighbors)


plt.figure()
# plt.subplot(1, 3, i + 1, aspect=1)

# Fit the method's model
kda.fit(X_train, y_train)

# Fit a nearest neighbor classifier on the embedded training set
knn.fit(kda.transform(X_train), y_train)

# Compute the nearest neighbor accuracy on the embedded test set
acc_knn = knn.score(kda.transform(X_test), y_test)

# Embed the data set in 2 dimensions using the fitted model
X_embedded = kda.transform(X)

# Plot the projected points and show the evaluation score
plt.scatter(X_embedded[:, 0], X_embedded[:, 1], c=y, s=30, cmap='Set1')
plt.title("{}, KNN (k={})\nTest accuracy = {:.2f}".format('kda',
                                                              n_neighbors,
                                                              acc_knn))
plt.show()
